'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _lodashObjectAssign = require('lodash/object/assign');

var _lodashObjectAssign2 = _interopRequireDefault(_lodashObjectAssign);

var _lodashCollectionForEach = require('lodash/collection/forEach');

var _lodashCollectionForEach2 = _interopRequireDefault(_lodashCollectionForEach);

var _lodashObjectHas = require('lodash/object/has');

var _lodashObjectHas2 = _interopRequireDefault(_lodashObjectHas);

var _lodashObjectMapValues = require('lodash/object/mapValues');

var _lodashObjectMapValues2 = _interopRequireDefault(_lodashObjectMapValues);

var _lodashObjectOmit = require('lodash/object/omit');

var _lodashObjectOmit2 = _interopRequireDefault(_lodashObjectOmit);

var _lodashObjectPick = require('lodash/object/pick');

var _lodashObjectPick2 = _interopRequireDefault(_lodashObjectPick);

var _stampit = require('stampit');

var _stampit2 = _interopRequireDefault(_stampit);

function isStamp(obj) {
  return typeof obj === 'function' && typeof obj.compose === 'function' && typeof obj.fixed === 'object';
}

function stripStamp(stamp) {
  delete stamp.create;
  delete stamp.methods;
  delete stamp.state;
  delete stamp.enclose;
  delete stamp['static'];

  return stamp;
}

var dupeFilter = function dupeFilter(prev, next, key, targ) {
  if (targ[key]) {
    throw new TypeError('Cannot mixin key `' + key + '` because it is provided by multiple sources.');
  }

  return next;
};

/**
 * Iterate through stamp methods, creating wrapper
 * functions for mixable React methods,
 * starting execution with last-in.
 *
 * @param {Object} dest Method destination
 * @param {Object} src New methods
 * @return {Object} An object of methods
 */
function wrapFunctions(targ, src) {
  var funcs = undefined;
  var mixability = {
    componentWillMount: 'many',
    componentDidMount: 'many',
    componentWillReceiveProps: 'many',
    componentWillUpdate: 'many',
    shouldComponentUpdate: 'once',
    componentDidUpdate: 'many',
    render: 'once',
    componentWillUnmount: 'many',
    getChildContext: 'many_merged' };

  funcs = (0, _lodashObjectMapValues2['default'])(src, function (val, key) {
    if (typeof val === 'function') {
      switch (mixability[key]) {
        case 'many':
          return function () {
            var res1 = targ[key] && targ[key].apply(this, arguments);
            var res2 = val.apply(this, arguments);

            return res2 || res1;
          };
        case 'many_merged':
          return function () {
            var res1 = targ[key] && targ[key].apply(this, arguments);
            var res2 = val.apply(this, arguments);

            if (res1) {
              return (0, _lodashObjectAssign2['default'])({}, res1, res2, dupeFilter);
            }

            return res2;
          };
        case 'once':
        default:
          if (!targ[key]) {
            return val;
          }

          throw new TypeError('Cannot mixin `' + key + '` because it has a unique constraint.');
      }
    }
  });

  return (0, _lodashObjectAssign2['default'])({}, targ, funcs);
}

/**
 * Process the static properties of a stamp and
 * combine the result with the passed in statics object.
 *
 * @param {Object} stamp A stamp
 * @param {Object} prev An object of past static properties
 * @return {Object} A processed object of static properties
 */
function extractStatics(targ, src) {
  var statics = (0, _lodashObjectAssign2['default'])({}, targ);
  var dupeCheck = ['propTypes', 'defaultProps'];

  (0, _lodashCollectionForEach2['default'])(src, function (val, key) {
    if (dupeCheck.indexOf(key) >= 0) {
      statics[key] = (0, _lodashObjectAssign2['default'])({}, statics[key], val, dupeFilter);
    } else if (typeof val === 'object') {
      statics[key] = (0, _lodashObjectAssign2['default'])({}, statics[key], val);
    } else {
      statics[key] = val;
    }
  });

  return statics;
}

/**
 * Take two or more stamps produced from react-stampit or
 * stampit and combine them to produce and return a new stamp.
 * Combining overrides properties with last-in priority.
 *
 * state - concatenative inheritance / cloning
 *
 * statics - concatenative inheritance / cloning
 *
 * methods - functional inheritance / closure prototypes
 *   - execute wrapped methods with first-in priority
 *
 * @param {...Function} stamp Two or more stamps.
 * @return {Function} A new stamp composed from arguments.
 */
function compose() {
  for (var _len = arguments.length, factories = Array(_len), _key = 0; _key < _len; _key++) {
    factories[_key] = arguments[_key];
  }

  var stamps = factories.slice(),
      state = { state: {} },
      methods = {},
      statics = {};

  if (isStamp(this)) {
    stamps.push(this);
  }

  (0, _lodashCollectionForEach2['default'])(stamps, function (stamp) {
    /* eslint-disable */
    stamp = !isStamp(stamp) ? rStampit(null, stamp) : stamp;
    /* eslint-enable */

    methods = wrapFunctions(methods, stamp.fixed.methods);
    statics = extractStatics(statics, stamp.fixed['static']);

    if (stamp.fixed.state && stamp.fixed.state.state) {
      state.state = (0, _lodashObjectAssign2['default'])({}, state.state, stamp.fixed.state.state, dupeFilter);
    }
  });

  return stripStamp((0, _lodashObjectAssign2['default'])((0, _stampit2['default'])().state(state).methods(methods)['static'](statics), { compose: compose }));
}

/**
 * Return a factory function (called a stamp) that will produce new
 * React components using the prototypes that are passed in or composed.
 *
 * @param {Object} React  The React library.
 * @param {Object} [props]  A map of property names and values specialized for React.
 * @return {Function} stamp A factory to produce React components using the given properties.
 * @return {Object} stamp.fixed An object map containing the fixed prototypes.
 */
function rStampit(React, props) {
  var react = React ? _stampit2['default'].convertConstructor(React.Component) : (0, _stampit2['default'])();
  var filtered = undefined,
      state = undefined,
      methods = undefined,
      statics = undefined;

  // shortcut for `convertConstructor`
  if (typeof props !== 'object' || Object.keys(props) === 0) {
    react.compose = compose;
    return stripStamp(react);
  }

  filtered = (0, _lodashObjectOmit2['default'])(props, ['state', 'statics']);
  statics = (0, _lodashObjectAssign2['default'])({}, props.statics, (0, _lodashObjectPick2['default'])(filtered, ['contextTypes', 'childContextTypes', 'propTypes', 'defaultProps']));
  methods = (0, _lodashObjectOmit2['default'])(filtered, function (val, key) {
    return (0, _lodashObjectHas2['default'])(statics, key);
  });
  state = props.state && { state: props.state };

  return stripStamp((0, _lodashObjectAssign2['default'])(react.state(state).methods(methods)['static'](statics), { compose: compose }));
}

exports['default'] = (0, _lodashObjectAssign2['default'])(rStampit, { compose: compose, isStamp: isStamp });
module.exports = exports['default'];